import math
import functools
from pathlib import Path
import png
import hilbert
import numpy as np
import polars as pl
from multiprocessing import Pool

hilbert_coords = functools.partial(hilbert.decode, num_dims=2, num_bits=16)

def convert_to_parquet(csv_path: Path, parquet_path: Path):
    if not csv_path.exists():
        print(f"no csv found at \"{csv_path}\"")
        return
    lf = pl.scan_csv(csv_path, schema={
        "saddr": pl.String,
        "rtt_us": pl.UInt64,
        "success": pl.UInt8
    })
    lf = lf.filter(pl.col("success") == 1)
    lf = lf.drop("success")
    lf = lf.with_columns(rtt_us=pl.col("rtt_us").clip(0, 0xFFFFFFFF).cast(pl.UInt32))
    lf = lf.with_columns(saddr=pl.col("saddr").str.split_exact(".", 3).struct.rename_fields(["a", "b", "c", "d"]))
    lf = lf.with_columns(saddr=pl.col("saddr").struct.field("a").cast(pl.UInt32) * 0x1000000 + pl.col("saddr").struct.field("b").cast(pl.UInt32) * 0x10000 + pl.col("saddr").struct.field("c").cast(pl.UInt32) * 0x100 + pl.col("saddr").struct.field("d").cast(pl.UInt32))
    lf = lf.with_columns(coords=pl.col("saddr").map_batches(hilbert_coords, pl.Array(pl.UInt16, 2), is_elementwise=True))
    lf = lf.with_columns(x=pl.col("coords").arr.get(0), y=pl.col("coords").arr.get(1))
    lf = lf.drop("coords")
    print(f"scanning csv \"{csv_path}\" into parquet \"{parquet_path}\"...", end=" ", flush=True)
    lf.sink_parquet(parquet_path)
    print("done.")

def write_tile(path: Path, rows: np.ndarray):
    path.parent.mkdir(exist_ok=True, parents=True)
    png.Writer(rows.shape[0], rows.shape[0], greyscale=False, alpha=False).write(path.open("wb"), rows)

def generate_tiles(parquet_path: Path, tiles_dir: Path, tile_size=256, processes=16):
    print(f"reading parquet \"{parquet_path}\"...", end=" ", flush=True)
    gdf = pl.read_parquet(parquet_path, columns=["x", "y", "rtt_us"])
    print("done.")
    
    channels = 3
    tile_size_by_channels = tile_size * channels
    tiles_per_side = int(math.sqrt(0x100000000)) // tile_size

    while True:
        # if tiles_per_side <= 16:
        z = int(math.log2(tiles_per_side))
        print(f"[{z=}] calculating colors...", end=" ", flush=True)
        df = gdf.with_columns(gb=0xFF - (pl.col("rtt_us") / 3000).round().clip(0, 0xFF).cast(pl.UInt8))
        df = df.drop("rtt_us")
        df = df.with_columns(r=0xFF, g=pl.col("gb"), b=pl.col("gb"))
        df = df.drop("gb")
        print("done.")
        total_size = tiles_per_side * tile_size
        print(f"[{z=}] creating image row data...", end=" ", flush=True)
        all_rows = np.zeros((total_size, total_size, channels), dtype = "uint8")
        all_rows[(df.get_column("y"), df.get_column("x"))] = df.select(["r", "g", "b"]).to_numpy()
        all_rows = all_rows.reshape(total_size, total_size * channels)
        print("done.")
        del df

        z_path = tiles_dir / f"{z}"
        z_path.mkdir(exist_ok=True, parents=True)
        print(f"[{z=}] creating individual tile data...", end=" ", flush=True)
        tile_data = [
            (z_path / f"{y}" / f"{x}.png", all_rows[
                y * tile_size : y * tile_size + tile_size,
                x * tile_size_by_channels : x * tile_size_by_channels + tile_size_by_channels
            ])
            for x in range(tiles_per_side)
            for y in range(tiles_per_side)
        ]
        print("done.")
        print(f"[{z=}] writing {tiles_per_side}x{tiles_per_side}={tiles_per_side * tiles_per_side} images...", end=" ", flush=True)
        with Pool(processes) as pool:
            pool.starmap(write_tile, tile_data)
        print("done.")
        del tile_data
        del all_rows

        if tiles_per_side == 1:
            break
        old_tiles_per_side = tiles_per_side
        tiles_per_side //= 2
        
        print(f"[{z=}] rescaling {len(gdf)} coords from {old_tiles_per_side}x{old_tiles_per_side} to {tiles_per_side}x{tiles_per_side} tiles...", end=" ", flush=True)
        new_gdf = gdf.with_columns(x=pl.col("x") // 2, y=pl.col("y") // 2)
        del gdf
        gdf = new_gdf.group_by(["x", "y"]).mean()
        print(f"done. {len(gdf)} coords remaining.")

def main():
    data_dir = Path("data")
    csv_path = data_dir / "full-scan.csv"
    parquet_path = data_dir / "full-scan.parquet"
    if not parquet_path.exists():
        print(f"no parquet file found at \"{parquet_path}\", generating now...")
        convert_to_parquet(csv_path, parquet_path)
    tiles_dir = Path("tiles")
    generate_tiles(parquet_path, tiles_dir)

if __name__ == "__main__":
    main()
